from typing import Dict, Type, Any, Union, Optional, Callable, Tuple, List
import time
import warnings
import functools
import datetime

import numpy as np
import torch
import sklearn
import sklearn.model_selection
import optuna

from . import path_helpers
from . import optimization
from . import linear_regression


class Autotuner_BaseEstimator:
    """
    A class for automatic hyperparameter tuning and training of a regression
    model.
    RH 2023
    
    Attributes:
        model_class (Type[sklearn.base.BaseEstimator]):
            A Scikit-Learn estimator class.
            Must have: \n
                * Method: ``fit(X, y)``
                * Method: ``predict_proba(X)`` (for classifiers) or ``predict(X)`` (for continuous regressors) \n
        X (np.ndarray):
            Input data.
            Shape: *(n_samples, n_features)*
        y (np.ndarray):
            Output data.
            Shape: *(n_samples,)*
        params (Dict[str, Dict[str, Any]]):
            A dictionary of hyperparameters with their names, types, and bounds.
        cv (Type[sklearn.model_selection._split.BaseCrossValidator]):
            A Scikit-Learn cross-validator class.
            Must have: \n
                * Call signature: ``idx_train, idx_test = next(self.cv.split(self.X, self.y))``
        fn_loss (Callable):
            Function to compute the loss.
            Must have: \n
                * Call signature: ``loss, loss_train, loss_test =
                  fn_loss(y_train_pred, y_test_pred, y_true_train, y_true_test,
                  sample_weight_train, sample_weight_test)`` \n
        params_linked (Dict):
            Dictionary of parameters that are linked. \n
            Each item in the dictionary is of the form: \n
                * ``{'param_name': (param_name_linked, lambda x: f(x))}`` \n
            Where ``'param_name_linked'`` matches the name of a parameter in
            ``params_dynamic`` and ``lambda x: f(x)`` is a function that takes
            the value of that parameter and returns the value to be used by
            ``'param_name'``.
        groups (Optional[np.ndarray]):
            Group labels for the samples used while splitting the dataset into
            train/test set. Only used if ``cv`` is a GroupShuffleSplit.
        n_jobs_optuna (int):
            Number of jobs for Optuna. Set to ``-1`` to use all cores.
            Note that some ``'solver'`` options are already parallelized (like
            ``'lbfgs'``). Set ``n_jobs_optuna`` to ``1`` for these solvers.
        n_startup (int):
            The number of startup trials for the optuna pruner and sampler.
        kwargs_convergence (Dict[str, Union[int, float]]):
            Convergence settings for the optimization. Includes: \n
                * ``'n_patience'`` (int): The number of trials to wait for
                  convergence before stopping the optimization.
                * ``'tol_frac'`` (float): The fractional tolerance for
                  convergence. After ``n_patience`` trials, the optimization
                  will stop if the loss has not improved by at least
                  ``tol_frac``.
                * ``'max_trials'`` (int): The maximum number of trials to run.
                * ``'max_duration'`` (int): The maximum duration of the
                  optimization in seconds. \n
        sample_weight (Optional[np.ndarray]):
            Weights for the samples, equal to ones_like(y) if None.
        catch_convergence_warnings (bool):
            If ``True``, ignore ConvergenceWarning during model fitting.
        verbose (bool):
            If ``True``, show progress bar and print running results.
        optuna_storage_url (Optional[str]):
            The path or URL to a SQL database. If ``None``, then the study is
            stored in memory. See optuna documentation for details. \n
        optuna_storage_name (Optional[str]):
            The name of the study in the storage. Should only be ``None`` if
            ``optuna_storage_url`` is ``None``, else it should be a string to a
            valid study name. 
        optuna_engine_kwargs (Optional[Dict[str, Any]]):
            Additional keyword arguments to pass to the SQLAlchemy engine.
        wandb_project (Optional[str]):
            The name of the Weights and Biases project to log to. If ``None``,
            then no logging is done. Assumes that a wandb.init() has already
            been called. \n


    Example:
        .. highlight:: python
        .. code-block:: python
    
        params = {
            'C':             {'type': 'real',        'kwargs': {'log': True, 'low': 1e-4, 'high': 1e4}},
            'penalty':       {'type': 'categorical', 'kwargs': {'choices': ['l1', 'l2']}},
        }
    """
    def __init__(
        self, 
        model_class: Type[sklearn.base.BaseEstimator], 
        X: Any, 
        y: Any, 
        params: Dict[str, Dict[str, Any]], 
        cv: Any, 
        fn_loss: Callable, 
        params_linked: Dict = {},
        groups: Optional[np.ndarray] = None,
        n_jobs_optuna: int = -1,
        n_startup: int = 15,
        kwargs_convergence = {
            'n_patience': 100,
            'tol_frac': 0.05,
            'max_trials': 350,
            'max_duration': 60*10,
            'use_internal_timer': True,
        }, 
        n_repeats: int = 1,
        fn_reduce_repeats: Callable = np.nanmedian,
        sample_weight: Optional[Any] = None, 
        catch_convergence_warnings: bool = True,
        verbose=True,
        optuna_storage_url: Optional[str] = None,
        optuna_storage_name: Optional[str] = None,
        optuna_engine_kwargs: Optional[Dict[str, Any]] = None,
        wandb_project: Optional[str] = None,
    ):
        """
        Initializes the AutotunerRegression with the given model class, parameters, data, and settings.
        """
        ## Set model variables
        self.X = X  ## shape (n_samples, n_features)
        self.y = y  ## shape (n_samples,)
        self.model_class = model_class  ## sklearn estimator class
        if isinstance(self.y, torch.Tensor):
            assert sample_weight is None, 'sample weights not supported for torch tensors.'
            self.sample_weight = None
        elif isinstance(self.y, np.ndarray):
            self.sample_weight = np.ones_like(self.y)
        self.cv = cv  ## sklearn cross-validator object with split method
        self.groups = groups  ## group labels for the samples used while splitting the dataset into train/test set

        ## Set optuna variables
        self.n_startup = n_startup
        self.params = params
        self.params_linked = params_linked
        self.n_jobs_optuna = n_jobs_optuna
        self.fn_loss = fn_loss
        self.catch_convergence_warnings = catch_convergence_warnings

        ## Set repeat variables
        self.n_repeats = n_repeats
        self.fn_reduce_repeats = fn_reduce_repeats


        self.kwargs_convergence = kwargs_convergence

        self.verbose = verbose
        self.optuna_storage_url = optuna_storage_url
        self.optuna_storage_name = optuna_storage_name
        self.optuna_engine_kwargs = optuna_engine_kwargs

        self.wandb_project = wandb_project

        # Initialize a convergence checker
        self.checker = optimization.Convergence_checker_optuna(verbose=verbose, **self.kwargs_convergence)

        # Initialize wandb callback
        if self.wandb_project is not None:
            import logging
            import sys
            optuna.logging.get_logger("optuna").addHandler(logging.StreamHandler(sys.stdout))
            config = {
                'model_class': self.model_class.func.__name__ if hasattr(self.model_class, 'func') else 'unknown',
                'params': self.params,
                'n_repeats': self.n_repeats,
                'n_startup': self.n_startup,
                'kwargs_convergence': self.kwargs_convergence,
                'n_jobs_optuna': self.n_jobs_optuna,
            }
            callback_wandb = optuna.integration.WeightsAndBiasesCallback(
                metric_name="loss",
                wandb_kwargs={
                    'project': self.wandb_project,
                    'config': config,
                },
                as_multirun=False,
            )
            ## Make a safe version of the callback by putting it in a try-except block
            def safe_callback(study, trial):
                try:
                    callback_wandb(study, trial)
                except Exception as e:
                    print(f'Error in wandb callback: {e}')
            self.callback_wandb = safe_callback
        else:
            self.callback_wandb = None

        # Set verbosity
        if int(self.verbose) <= 1:
            optuna.logging.set_verbosity(optuna.logging.WARNING)
        elif int(self.verbose) == 2:
            optuna.logging.set_verbosity(optuna.logging.INFO)
        elif int(self.verbose) > 2:
            optuna.logging.set_verbosity(optuna.logging.DEBUG)

        # Initialize an Optuna study
        if self.optuna_storage_url is not None:
            storage = optuna.storages.RDBStorage(
                url=self.optuna_storage_url,
                engine_kwargs=self.optuna_engine_kwargs,
            )
        else:
            storage = None
        self.study = optuna.create_study(
            direction="minimize", 
            pruner=optuna.pruners.MedianPruner(n_startup_trials=self.n_startup), 
            sampler=optuna.samplers.TPESampler(n_startup_trials=self.n_startup),
            study_name='Autotuner' if self.optuna_storage_name is None else self.optuna_storage_name,
            storage=storage,
            load_if_exists=True,
        )
        
        # Initialize variables to store loss and best model
        self.loss_running_train = []
        self.loss_running_test = []
        self.loss_running = []
        self.loss_repeats_train = []
        self.loss_repeats_test = []
        self.loss_repeats = []
        self.params_running = []
        self.model_best = None
        self.loss_best = np.inf
        self.params_best = None

    def _objective(self, trial: optuna.trial.Trial) -> float:
        """
        Define the objective function for Optuna to optimize.

        Args:
            trial (optuna.trial.Trial): 
                An Optuna trial object.

        Returns:
            (float): 
                loss (float):
                    The score of the model.
        """
        # Make a lookup table for the suggest methods
        LUT_suggest = {
            'categorical': trial.suggest_categorical,
            'real': trial.suggest_float,
            'int': trial.suggest_int,
            'discrete_uniform': trial.suggest_discrete_uniform,
            'loguniform': trial.suggest_loguniform,
            'uniform': trial.suggest_uniform,
        }

        # Suggest hyperparameters using optuna
        kwargs_model = {}
        for name, config in self.params.items():
            kwargs_model[name] = LUT_suggest[config['type']](name, **config['kwargs'])
        ## Add linked parameters
        for param, (param_linked, fn) in self.params_linked.items():
            kwargs_model[param] = fn(kwargs_model[param_linked])

        # Train the model
        loss_train_all, loss_test_all, loss_all = [], [], []
        for ii in range(self.n_repeats):
            # Split the data
            idx_train, idx_test = next(self.cv.split(self.X, self.y, **({'groups': self.groups} if self.groups is not None else {}) ))
            X_train, y_train_true, X_test, y_test_true = self.X[idx_train], self.y[idx_train], self.X[idx_test], self.y[idx_test]
            if self.sample_weight is not None:
                sample_weight_train, sample_weight_test = self.sample_weight[idx_train], self.sample_weight[idx_test]
            else:
                sample_weight_train, sample_weight_test = None, None

            # Initialize the model with the suggested hyperparameters
            model = self.model_class(**kwargs_model)

            # Train the model
            ## Turn off ConvergenceWarning
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=sklearn.exceptions.ConvergenceWarning) if self.catch_convergence_warnings else None
                model.fit(X_train, y_train_true)

            # Transform the training data
            if hasattr(model, 'predict_proba'):
                y_train_pred = model.predict_proba(X_train)
                y_test_pred = model.predict_proba(X_test)
            elif hasattr(model, 'predict'):
                y_train_pred = model.predict(X_train)
                y_test_pred = model.predict(X_test)
            else:
                raise ValueError('Model must have either a predict_proba or predict method.')

            # Evaluate the model using the scoring method
            loss, loss_train, loss_test = self.fn_loss(
                y_train_pred=y_train_pred, 
                y_test_pred=y_test_pred,
                y_train_true=y_train_true,
                y_test_true=y_test_true,
                sample_weight_train=sample_weight_train,
                sample_weight_test=sample_weight_test,
            )

            loss_train_all.append(loss_train)
            loss_test_all.append(loss_test)
            loss_all.append(loss)

        # Reduce the loss
        if len(loss_train_all) == 1:
            loss_train, loss_test, loss = loss_train_all[0], loss_test_all[0], loss_all[0]        
        else:
            if isinstance(loss_train_all[0], (np.ndarray, np.generic, float, int)):
                stack = np.stack
            elif isinstance(loss_train_all[0], torch.Tensor):
                stack = torch.stack
            elif loss_train_all[0] is None:
                stack = lambda x: x
            else:
                raise ValueError(f'loss_train_all[0] must be either np.ndarray, np.generic, torch.Tensor, or None. Got {type(loss_train_all[0])}')
            loss_train, loss_test, loss = (self.fn_reduce_repeats(stack(l)) for l in [loss_train_all, loss_test_all, loss_all])


        # Save the running loss
        self.loss_running_train.append(loss_train)
        self.loss_running_test.append(loss_test)
        self.loss_running.append(loss)
        self.loss_repeats_train.append(loss_train_all)
        self.loss_repeats_test.append(loss_test_all)
        self.loss_repeats.append(loss_all)
        self.params_running.append(kwargs_model)

        # Update the bests
        loss = np.nan if loss is None else loss
        if loss < self.loss_best:
            self.loss_best = loss
            self.model_best = model
            self.params_best = kwargs_model

        # Log the trial results in optuna: losses
        trial.set_user_attr('loss', loss)
        trial.set_user_attr('loss_train', loss_train)
        trial.set_user_attr('loss_test', loss_test)

        return loss

    def fit(self) -> Union[sklearn.base.BaseEstimator, Optional[Dict[str, Any]]]:
        """
        Fit and tune the hyperparameters and train the model.

        Returns:
            (Union[sklearn.base.BaseEstimator, Optional[Dict[str, Any]]): 
                best_model (sklearn.base.BaseEstimator):
                    The best estimator obtained from hyperparameter tuning.
                best_params (Optional[Dict[str, Any]]):
                    The best parameters obtained from hyperparameter tuning.
        """
        # Optimize the study
        callbacks = [self.checker.check] + ([self.callback_wandb] if self.callback_wandb is not None else [])
        self.study.optimize(
            self._objective, 
            n_jobs=self.n_jobs_optuna, 
            callbacks=callbacks, 
            n_trials=self.kwargs_convergence['max_trials'],
            show_progress_bar=self.verbose,
        )

        # Print reason for convergence
        if self.verbose > 0:
            print(f'Convergence reason: {self.checker.reason_converged}')

        # Retrieve the best parameters and the best model
        self.best_params = self.study.best_params
        self.model_best = self.model_class(**self.best_params)
        
        # Train the model on the full data set
        self.model_best.fit(self.X, self.y)

        # End wandb run
        if self.wandb_project is not None:
            import wandb
            wandb.finish()

        return self.model_best, self.best_params
    
    def save_model(
        self,
        filepath: Optional[str]=None,
        allow_overwrite: bool=False,
    ):
        """
        Uses ONNX to save the best model as a binary file.

        Args:
            filepath (str): 
                The path to save the model to.
                If None, then the model will not be saved.
            allow_overwrite (bool):
                Whether to allow overwriting of existing files.

        Returns:
            (onnx.ModelProto):
                The ONNX model.
        """
        import datetime
        try:
            import onnx
            import skl2onnx
        except ImportError as e:
            raise ImportError(f'You need to (pip) install onnx and skl2onnx to use this method. {e}')
        
        ## Make sure we have what we need
        assert self.model_best is not None, 'You need to fit the model first.'

        # Convert the model to ONNX format
        ## Prepare initial types
        initial_types = skl2onnx.common.data_types.guess_data_type(self.X)[0][1]
        initial_types.shape = tuple([None] + list(initial_types.shape[1:]))
        initial_types = [('input', initial_types)]
        ### Convert the model
        model_onnx = skl2onnx.convert_sklearn(
            model=self.model_best,
            name=str(self.model_best.__class__),
            initial_types=initial_types,
            doc_string=f"Created by Autotuner. Saved on {datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')}.",
        )
        
        # Save the model
        if filepath is not None:
            path_helpers.prepare_filepath_for_saving(filepath=filepath, mkdir=True, allow_overwrite=allow_overwrite)
            onnx.save(
                proto=model_onnx,
                f=filepath,
            )
        return model_onnx
    
    def plot_param_curve(
        self,
        param='alpha',
        xscale='linear',
        yscale='linear',
        jitter=0.01,
    ):
        """
        Makes a scatter plot of a selected values vs loss values.

        Args:
            param (str):
                The parameter to plot.
            xscale (str):
                The scale of the x-axis. Either ``'linear'`` or ``'log'``.
        """
        def prep(x):
            """to numpy"""
            if isinstance(x[0], torch.Tensor):
                return torch.stack(x).detach().cpu().numpy()
            else:
                return np.stack(x)

        import matplotlib.pyplot as plt
        results = {
            param: prep([t[param] for t in self.params_running]), 
            'value': prep(self.loss_running), 
            'loss_train': prep(self.loss_running_train), 
            'loss_test': prep(self.loss_running_test),
        }
       
        fig = plt.figure()
        isnumeric = np.issubdtype(results[param].dtype, np.number)
        n_vals = len(results[param])
        range_vals = np.ptp(results['value'])
        jitter_scaling = np.ptp(results[param]) if isnumeric else n_vals
        jitter_vals = np.random.uniform(-jitter, jitter, size=n_vals) * jitter * jitter_scaling
        xaxis = results[param] if isnumeric else np.unique(results[param], return_inverse=True)[1]
        xaxis_unique, xaxis_counts = np.unique(xaxis, return_counts=True)
        results_means = np.array([(u, np.mean(results['value'][xaxis == u])) for u in xaxis_unique])
        scatter = functools.partial(
            plt.scatter, 
            x=xaxis + jitter_vals,
            alpha=0.3, 
            s=10
        )
        scatter(y=results['loss_train'], color='blue')
        scatter(y=results['loss_test'], color='orange')
        scatter(y=results['value'], color='k')
        plt.scatter(results_means[:,0], results_means[:,1], s=70, color='k') if any(xaxis_counts > 1) else None
        plt.scatter(x=xaxis[np.argmin(results['value'])], y=np.min(results['value']), color='r', s=100, alpha=1)

        plt.xlabel(param)
        plt.ylabel('loss')
        if isnumeric:
            plt.xscale(xscale)
            plt.yscale(yscale)
        else:
            plt.xticks(xaxis, results[param], rotation=90)
            plt.xlim(-1, len(xaxis_unique))
            plt.ylim(np.min(results['value']) + (range_vals * -0.1), np.max(results['value']) + (range_vals * 0.1))
        plt.legend([
            'loss_train',
            'loss_test',
            'loss_withPenalty',
            'loss_withPenalty_mean',
            'best model',
        ])

        return fig

    def to(self, device):
        if isinstance(self.model_best, linear_regression.LinearRegression_sk):
            self.model_best.to(device)
        return self
    def cpu(self):
        return self.to('cpu')
    def numpy(self):
        if isinstance(self.model_best, linear_regression.LinearRegression_sk):
            self.model_best.numpy()
        return self

class LossFunction_CrossEntropy_CV():
    """
    Calculates the cross-entropy loss of a classifier using cross-validation. 
    RH 2023

    Args:
        penalty_testTrainRatio (float): 
            The amount of penalty for the test loss to the train loss. 
            Penalty is applied with formula: 
            ``loss = loss_test_or_train * ((loss_test / loss_train) ** penalty_testTrainRatio)``.
        labels (Optional[Union[List, np.ndarray]]): 
            A list or ndarray of labels. 
            Shape: *(n_samples,)*.
        test_or_train (str): 
            A string indicating whether to apply the penalty to the test or
            train loss.
            It should be either ``'test'`` or ``'train'``. 
    """
    def __init__(
        self,
        penalty_testTrainRatio: float = 1.0,
        labels: Optional[Union[List, np.ndarray]] = None,
        test_or_train: str = 'test',
    ) -> None:
        """
        Initializes the LossFunctionCrossEntropyCV with the given penalty, labels, and test_or_train setting.
        """
        self.labels = labels
        self.penalty_testTrainRatio = penalty_testTrainRatio
        ## Set the penalty function
        if test_or_train == 'test':
            self.fn_penalty_testTrainRatio = lambda test, train: test * ((test  / train) ** self.penalty_testTrainRatio)
        elif test_or_train == 'train':
            self.fn_penalty_testTrainRatio = lambda test, train: train * ((train / test) ** self.penalty_testTrainRatio)
        else:
            raise ValueError('test_or_train must be either "test" or "train".')

    
    def __call__(
        self,
        y_train_pred: np.ndarray, 
        y_test_pred: np.ndarray,
        y_train_true: np.ndarray,
        y_test_true: np.ndarray,
        sample_weight_train: Optional[List[float]] = None,
        sample_weight_test: Optional[List[float]] = None,
    ):
        """
        Calculates the cross-entropy loss using cross-validation.

        Args:
            y_train_pred (np.ndarray): 
                Predicted output probabilities for the training set. (shape:
                *(n_samples,)*)
            y_test_pred (np.ndarray): 
                Predicted output probabilities for the test set. (shape:
                *(n_samples,)*)
            y_train_true (np.ndarray): 
                True output probabilities for the training set. (shape:
                *(n_samples,)*)
            y_test_true (np.ndarray): 
                True output probabilities for the test set. (shape:
                *(n_samples,)*)
            sample_weight_train (Optional[List[float]]): 
                Weights assigned to each training sample. 
            sample_weight_test (Optional[List[float]]): 
                Weights assigned to each test sample.

        Returns:
            (tuple): tuple containing:
                loss (float): 
                    The calculated loss after applying the penalty.
                loss_train (float): 
                    The cross-entropy loss of the training set.
                loss_test (float): 
                    The cross-entropy loss of the test set.
        """
        # Calculate the cross-entropy loss using cross-validation.
        from sklearn.metrics import log_loss
        loss_train = log_loss(y_train_true, y_train_pred, sample_weight=sample_weight_train, labels=self.labels)
        loss_test =  log_loss(y_test_true,  y_test_pred,  sample_weight=sample_weight_test,  labels=self.labels)
        loss = self.fn_penalty_testTrainRatio(loss_test, loss_train)

        return loss, loss_train, loss_test


class LossFunction_MSE_CV():
    """
    Calculates the mean-squared error loss of a model using
    cross-validation. 
    RH 2023

    Args:
        penalty_testTrainRatio (float): 
            The amount of penalty for the test loss to the train loss. 
            Penalty is applied with formula: 
            ``loss = loss_test_or_train * ((loss_test / loss_train) ** penalty_testTrainRatio)``.
        test_or_train (str): 
            A string indicating whether to apply the penalty to the test or
            train loss.
            It should be either ``'test'`` or ``'train'``. 
    """
    def __init__(
        self,
        penalty_testTrainRatio: float = 1.0,
        test_or_train: str = 'test',
    ) -> None:
        """
        Initializes the class with the given penalty, and test_or_train setting.
        """
        self.penalty_testTrainRatio = penalty_testTrainRatio
        ## Set the penalty function
        if test_or_train == 'test':
            self.fn_penalty_testTrainRatio = lambda test, train: test * ((test  / train) ** self.penalty_testTrainRatio)
        elif test_or_train == 'train':
            self.fn_penalty_testTrainRatio = lambda test, train: train * ((train / test) ** self.penalty_testTrainRatio)
        else:
            raise ValueError('test_or_train must be either "test" or "train".')

    
    def __call__(
        self,
        y_train_pred: np.ndarray, 
        y_test_pred: np.ndarray,
        y_train_true: np.ndarray,
        y_test_true: np.ndarray,
        sample_weight_train: Optional[List[float]] = None,
        sample_weight_test: Optional[List[float]] = None,
    ):
        """
        Calculates the mean-squared error loss using cross-validation.

        Args:
            y_train_pred (np.ndarray): 
                Predicted output data for the training set. (shape:
                *(n_samples,)*)
            y_test_pred (np.ndarray): 
                Predicted output data for the test set. (shape: *(n_samples,)*)
            y_train_true (np.ndarray): 
                True output data for the training set. (shape: *(n_samples,)*)
            y_test_true (np.ndarray): 
                True output data for the test set. (shape: *(n_samples,)*)
            sample_weight_train (Optional[List[float]]): 
                Weights assigned to each training sample. 
            sample_weight_test (Optional[List[float]]): 
                Weights assigned to each test sample.

        Returns:
            (tuple): tuple containing:
                loss (float): 
                    The calculated loss after applying the penalty.
                loss_train (float): 
                    The mean-squared error loss of the training set.
                loss_test (float): 
                    The mean-squared error loss of the test set.
        """
        # Normalize the y values such that the variance of the true values is 1.
        y_train_pred = y_train_pred / y_train_true.std()
        y_test_pred = y_test_pred / y_test_true.std()
        y_train_true = y_train_true / y_train_true.std()
        y_test_true = y_test_true / y_test_true.std()

        # Calculate the mean-squared error loss using cross-validation.
        if isinstance(y_train_pred, np.ndarray):
            from sklearn.metrics import mean_squared_error
            loss_train = mean_squared_error(y_train_true, y_train_pred, sample_weight=sample_weight_train)
            loss_test =  mean_squared_error(y_test_true,  y_test_pred,  sample_weight=sample_weight_test)
            loss = self.fn_penalty_testTrainRatio(loss_test, loss_train)
        elif isinstance(y_train_pred, torch.Tensor):
            from torch.nn.functional import mse_loss
            assert (sample_weight_test is None) and (sample_weight_train is None), 'sample weights not supported for torch tensors.'
            loss_train = mse_loss(y_train_pred, y_train_true, reduction='mean').item()
            loss_test =  mse_loss(y_test_pred,  y_test_true,  reduction='mean').item()
            loss = self.fn_penalty_testTrainRatio(loss_test, loss_train)
        else:
            raise ValueError(f'Expected y_train_pred to be of type np.ndarray or torch.Tensor, but got type {type(y_train_pred)}.')
        
        return loss, loss_train, loss_test


class LossFunction_EVR_CV():
    """
    Calculates the explained variance ratio loss of a model using
    cross-validation. Output loss is 1 - EVR.
    RH 2024

    Args:
        penalty_testTrainRatio (float): 
            The amount of penalty for the test loss to the train loss. 
            Penalty is applied with formula: 
            ``loss = loss_test_or_train * ((loss_test / loss_train) ** penalty_testTrainRatio)``.
        test_or_train (str): 
            A string indicating whether to apply the penalty to the test or
            train loss.
            It should be either ``'test'`` or ``'train'``. 
    """
    def __init__(
        self,
        penalty_testTrainRatio: float = 1.0,
        test_or_train: str = 'test',
    ) -> None:
        """
        Initializes the class with the given penalty, and test_or_train setting.
        """
        self.penalty_testTrainRatio = penalty_testTrainRatio
        ## Set the penalty function
        if test_or_train == 'test':
            self.fn_penalty_testTrainRatio = lambda test, train: test * ((test  / train) ** self.penalty_testTrainRatio)
        elif test_or_train == 'train':
            self.fn_penalty_testTrainRatio = lambda test, train: train * ((train / test) ** self.penalty_testTrainRatio)
        else:
            raise ValueError('test_or_train must be either "test" or "train".')

    def explainable_variance_ratio(self, v1, v2, sample_weight=None):
        if isinstance(v1, torch.Tensor):
            v1 = v1 - torch.nanmean(v1, dim=0)
            v2 = v2 - torch.nanmean(v2, dim=0)

            v1_orth = v1 - (torch.nansum(v1 * v2, dim=0) / torch.nansum(v2 * v2, dim=0) )*v2

            v1_var = torch.var(v1, dim=0)
            EVR = 1 - (torch.var(v1_orth, dim=0) / v1_var)
            if sample_weight is not None:
                EVR = EVR * (sample_weight / torch.mean(sample_weight))

            EVR_total_weighted = torch.nansum(v1_var * EVR) / torch.sum(v1_var)
            return EVR_total_weighted
        elif isinstance(v1, np.ndarray):
            v1 = v1 - np.nanmean(v1, axis=0)
            v2 = v2 - np.nanmean(v2, axis=0)

            v1_orth = v1 - (np.nansum(v1 * v2, axis=0) / np.nansum(v2 * v2, axis=0) )*v2

            v1_var = np.var(v1, axis=0)
            EVR = 1 - (np.var(v1_orth, axis=0) / v1_var)
            if sample_weight is not None:
                EVR = EVR * (sample_weight / np.mean(sample_weight))

            EVR_total_weighted = np.nansum(v1_var * EVR) / np.sum(v1_var)
            return EVR_total_weighted
        else:
            raise ValueError(f'Expected v1 to be of type np.ndarray or torch.Tensor, but got type {type(v1)}.')

    def __call__(
        self,
        y_train_pred: np.ndarray, 
        y_test_pred: np.ndarray,
        y_train_true: np.ndarray,
        y_test_true: np.ndarray,
        sample_weight_train: Optional[List[float]] = None,
        sample_weight_test: Optional[List[float]] = None,
    ):
        """
        Calculates the explained variance ratio loss using cross-validation.
        Args:
            y_train_pred (np.ndarray): 
                Predicted output data for the training set. (shape:
                *(n_samples,)*)
            y_test_pred (np.ndarray): 
                Predicted output data for the test set. (shape: *(n_samples,)*)
            y_train_true (np.ndarray): 
                True output data for the training set. (shape: *(n_samples,)*)
            y_test_true (np.ndarray): 
                True output data for the test set. (shape: *(n_samples,)*)
            sample_weight_train (Optional[List[float]]): 
                Weights assigned to each training sample. 
            sample_weight_test (Optional[List[float]]): 
                Weights assigned to each test sample.

        Returns:
            (tuple): tuple containing:
                loss (float): 
                    The calculated loss after applying the penalty.
                loss_train (float): 
                    1 - explained variance ratio loss of the training set.
                loss_test (float): 
                    1 - explained variance ratio loss of the test set.
        """
        # Calculate 1 - explained variance ratio loss using cross-validation.
        
        loss_train = 1 - self.explainable_variance_ratio(y_train_true, y_train_pred, sample_weight_train)
        loss_test =  1 - self.explainable_variance_ratio(y_test_true, y_test_pred, sample_weight_test)
        loss = self.fn_penalty_testTrainRatio(loss_test, loss_train)
        
        return loss, loss_train, loss_test    


class Auto_Classifier(Autotuner_BaseEstimator):
    """
    Implements automatic hyperparameter tuning for a user defined classification model.
    RH 2023

    Args:
        X (np.ndarray):
            Training data. (shape: *(n_samples, n_features)*)
        y (np.ndarray):
            Target variable. (shape: *(n_samples, n_features)*)
        Model (class):
            A class to be used as the model. 
            Must have: \n
                * Method: ``fit(X, y)``
                * Method: ``predict_proba(X)`` (for classifiers) \n            
        params (Dict):
            Dictionary of model class initialization parameters. 
            For each item in the dictionary if item is: \n
                * ``list``: The parameter is tuned. If the values are numbers,
                  then the list wil be the bounds [low, high] to search over. If
                  the values are strings, then the list will be the categorical
                  values to search over. Note that the type of the parameter
                  (int, float, str) will be used to determine the type of the
                  optuna suggest.
                * **not** a ``list``: The parameter is fixed to the given value. \n
        params_log (Dict):
            Dictionary of parameters that vary logarithmically: \n
                * If True: parameter is set to logaritmically varying. \n
                * If False: parameter is set to linearly varying. \n
                * If not specified: parameter is set to logaritmically varying.
                  \n
        params_linked (Dict):
            Dictionary of parameters that are linked. \n
            Each item in the dictionary is of the form: \n
                * ``{'param_name': (param_name_linked, lambda x: f(x))}`` \n
            Where ``'param_name_linked'`` matches the name of a parameter in
            ``params_dynamic`` and ``lambda x: f(x)`` is a function that takes
            the value of that parameter and returns the value to be used by
            ``'param_name'``.
        n_startup (int):
            Number of startup trials.
        kwargs_convergence (Dict[str, Union[int, float]]):
            Convergence settings for the optimization. Includes: \n
                * ``'n_patience'`` (int): The number of trials to wait for
                  convergence before stopping the optimization.
                * ``'tol_frac'`` (float): The fractional tolerance for
                  convergence. After ``n_patience`` trials, the optimization
                  will stop if the loss has not improved by at least
                  ``tol_frac``.
                * ``'max_trials'`` (int): The maximum number of trials to run.
                * ``'max_duration'`` (int): The maximum duration of the
                  optimization in seconds. \n
        n_repeats (int):
            Number of repetitions for each trial.
        fn_reduce_repeats (Callable):
            Function to aggregate results from repetitions. (Default:
            `np.median`)
        n_jobs_optuna (int):
            Number of jobs for Optuna. Set to ``-1`` to use all cores.
            Note that some ``'solver'`` options are already parallelized (like
            ``'lbfgs'``). Set ``n_jobs_optuna`` to ``1`` for these solvers.
        penalty_testTrainRatio (float):
            Penalty ratio for test and train. 
        sample_weight (Optional[List[float]]):
            Sample weights. Applied only during cross-validation train/test
            accuracy, not model training. \n
            See `LogisticRegression
            <https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LogisticRegression.html>`_
            for more details.
        cv (Optional[sklearn.model_selection._split.BaseCrossValidator]):
            An Scikit-Learn cross-validator class.
            If not ``None``, then must have: \n
                * Call signature: ``idx_train, idx_test =
                  next(self.cv.split(self.X, self.y))`` \n
            If ``None``, then a StratifiedShuffleSplit cross-validator will be
            used.
        groups (Optional[np.ndarray]):
            Group labels for the samples used while splitting the dataset into
            train/test set. \n
        test_size (float):
            Test set ratio.
            Only used if ``cv`` is ``None``.
        loss_function (Optional[Callable]):
            A custom loss function. If ``None``, then the default loss function
            is used (CrossEntropy_CV).
        verbose (bool):
            Whether to print progress messages.
        optuna_storage_url (Optional[str]):
            The path or URL to a SQL database. If ``None``, then the study is
            stored in memory. See optuna documentation for details. \n
        optuna_storage_name (Optional[str]):
            The name of the study in the storage. Should only be ``None`` if
            ``optuna_storage_url`` is ``None``, else it should be a string to a
            valid study name. 
        optuna_engine_kwargs (Optional[Dict[str, Any]]):
            Additional keyword arguments to pass to the SQLAlchemy engine.
        wandb_project (Optional[str]):
            The name of the Weights and Biases project to log to. If ``None``,
            then no logging is done. Assumes that a wandb.init() has already
            been called. \n

    Demo:
        .. code-block:: python

            ## Initialize sklearn LogisticRegression class with NO TUNING. All
            parameters are fixed.
            autoclassifier = Auto_Classifier(
                X, 
                y, 
                params_dynamic={},
                params_static={
                    'C': 1e-14,
                    'penalty': 'l2',
                    'solver': 'lbfgs',
                },
            )

            ## Initialize with TUNING 'C', 'penalty', and 'l1_ratio'. 'solver' is fixed.
            autoclassifier = Auto_Classifier(
                X,
                y,
                params_dynamic={
                    'C': [1e-14, 1e3],
                    'penalty': ['l1', 'l2', 'elasticnet'],
                    'l1_ratio': [0.0, 1.0],
                },
                params_static={
                    'solver': 'lbfgs',
                },
                params_log={
                    'C': True,
                    'l1_ratio': False,
                },
            )
    """
    def __init__(
        self,
        X: np.ndarray,
        y: np.ndarray,
        Model: Callable,
        params_dynamic: Dict = {
            'alpha': [1e-1, 1e5],
        },
        params_static: Dict = {
            'fit_intercept': True,
            'max_iter': 1000,
            'tol': 0.0001,
            'solver': 'auto',
        },
        params_log: Dict = {
            'C': True,
            'max_iter': False,
        },
        params_linked: Dict = {},
        n_startup: int = 15,
        kwargs_convergence: Dict = {
            'n_patience': 50,
            'tol_frac': 0.05,
            'max_trials': 150,
            'max_duration': 60*10,
        }, 
        n_repeats: int = 1,
        fn_reduce_repeats: Callable = np.median,
        n_jobs_optuna: int = 1,
        penalty_testTrainRatio: float = 1.0,
        sample_weight: Optional[List[float]] = None,
        cv: Optional[sklearn.model_selection._split.BaseCrossValidator] = None,
        groups: Optional[np.ndarray] = None,
        test_size: float = 0.3,
        loss_function: Optional[Callable] = None,
        verbose: bool = True,
        optuna_storage_url: Optional[str] = None,
        optuna_storage_name: Optional[str] = None,
        optuna_engine_kwargs: Optional[Dict[str, Any]] = None,
        wandb_project: Optional[str] = None,
    ) -> None:
        """
        Initializes Auto_Classifier with the given parameters and data.
        """
        assert y.ndim == 1, f'y.ndim must be 1. Found {y.ndim}.'

        ## Prepare sample weights
        self.sample_weight = sklearn.utils.class_weight.compute_sample_weight(
            class_weight=sample_weight, 
            y=y,
        )

        ## Prepare the loss function
        if loss_function is None:
            self.fn_loss = LossFunction_CrossEntropy_CV(
                penalty_testTrainRatio=penalty_testTrainRatio,
                labels=y,
                test_or_train='test',
            )
        else:
            self.fn_loss = loss_function

        ## Prepare the cross-validation
        self.cv = sklearn.model_selection.StratifiedShuffleSplit(
            n_splits=1,
            test_size=test_size,
        ) if cv is None else cv

        ## Prepare dynamic parameters for optuna
        params = _infer_params_types(params_dynamic, params_log)

        ## Prepare the model class
        self.model_class = functools.partial(
            Model,
            **params_static,
        )

        ## Initialize the Autotuner superclass
        super().__init__(
            model_class=self.model_class,
            X=X,
            y=y,
            params=params,
            params_linked=params_linked,
            n_startup=n_startup,
            kwargs_convergence=kwargs_convergence,
            n_repeats=n_repeats,
            fn_reduce_repeats=fn_reduce_repeats,
            n_jobs_optuna=n_jobs_optuna,
            cv=self.cv,
            groups=groups,
            fn_loss=self.fn_loss,
            catch_convergence_warnings=True,
            verbose=verbose,
            optuna_storage_url=optuna_storage_url,
            optuna_storage_name=optuna_storage_name,
            optuna_engine_kwargs=optuna_engine_kwargs,
            wandb_project=wandb_project,
        )

    def evaluate_model(
        self, 
        model: Optional[object] = None,
        X: Optional[np.ndarray] = None, 
        y: Optional[np.ndarray] = None,
        sample_weight: Optional[List[float]] = None,
    ) -> Tuple[float, np.array]:
        """
        Evaluates the given model on the given data. Makes label predictions,
        then computes the accuracy and confusion matrix.

        Args:
            model (Optional[object]):
                The model to evaluate.
                If None, then self.model_best is used.
            X (np.ndarray):
                The data to evaluate on.
                If None, then self.X is used.
            y (np.ndarray):
                The labels to evaluate on.
                If None, then self.y is used.
            sample_weight (List[float]):
                The sample weights to evaluate on.
                If None, then self.sample_weight is used.

        Returns:
            (tuple): Tuple containing:
                accuracy (float):
                    The accuracy of the model on the given data.
                confusion_matrix (np.array):
                    The confusion matrix of the model on the given data.
        """
        model = self.model_best if model is None else model
        X = self.X if X is None else X
        y = self.y if y is None else y
        sample_weight = self.sample_weight if sample_weight is None else sample_weight

        y_pred = model.predict(X)
        
        accuracy = sklearn.metrics.accuracy_score(
            y_true=y,
            y_pred=y_pred,
            sample_weight=sample_weight,
            normalize=True,
        )

        confusion_matrix = sklearn.metrics.confusion_matrix(
            y_true=y,
            y_pred=y_pred,
            sample_weight=sample_weight,
            labels=self.classes,
            normalize='true',
        ).T

        return accuracy, confusion_matrix
    

class Auto_Regression(Autotuner_BaseEstimator):
    """
    Implements automatic hyperparameter tuning for a user defined classification model.
    RH 2024

    Args:
        X (np.ndarray):
            Training data. (shape: *(n_samples, n_features)*)
        y (np.ndarray):
            Target variable. (shape: *(n_samples, n_features)*)
        Model (class):
            A class to be used as the model. 
            Must have: \n
                * Method: ``fit(X, y)``
                * Method: ``predict(X)`` \n
        params (Dict):
            Dictionary of model class initialization parameters. 
            For each item in the dictionary if item is: \n
                * ``list``: The parameter is tuned. If the values are numbers,
                  then the list wil be the bounds [low, high] to search over. If
                  the values are strings, then the list will be the categorical
                  values to search over. Note that the type of the parameter
                  (int, float, str) will be used to determine the type of the
                  optuna suggest.
                * **not** a ``list``: The parameter is fixed to the given value. \n
        params_log (Dict):
            Dictionary of parameters that vary logarithmically: \n
                * If True: parameter is set to logaritmically varying. \n
                * If False: parameter is set to linearly varying. \n
                * If not specified: parameter is set to logaritmically varying.
                  \n
        params_linked (Dict):
            Dictionary of parameters that are linked. \n
            Each item in the dictionary is of the form: \n
                * ``{'param_name': (param_name_linked, lambda x: f(x))}`` \n
            Where ``'param_name_linked'`` matches the name of a parameter in
            ``params_dynamic`` and ``lambda x: f(x)`` is a function that takes
            the value of that parameter and returns the value to be used by
            ``'param_name'``.
        n_startup (int):
            Number of startup trials.
        kwargs_convergence (Dict[str, Union[int, float]]):
            Convergence settings for the optimization. Includes: \n
                * ``'n_patience'`` (int): The number of trials to wait for
                  convergence before stopping the optimization.
                * ``'tol_frac'`` (float): The fractional tolerance for
                  convergence. After ``n_patience`` trials, the optimization
                  will stop if the loss has not improved by at least
                  ``tol_frac``.
                * ``'max_trials'`` (int): The maximum number of trials to run.
                * ``'max_duration'`` (int): The maximum duration of the
                  optimization in seconds. \n
        n_repeats (int):
            Number of repetitions for each trial.
        fn_reduce_repeats (Callable):
            Function to aggregate results from repetitions. (Default:
            `np.median`)
        n_jobs_optuna (int):
            Number of jobs for Optuna. Set to ``-1`` to use all cores.
            Note that some ``'solver'`` options are already parallelized (like
            ``'lbfgs'``). Set ``n_jobs_optuna`` to ``1`` for these solvers.
        penalty_testTrainRatio (float):
            Penalty ratio for test and train. 
        sample_weight (Optional[List[float]]):
            Sample weights. Applied only during cross-validation train/test
            accuracy, not model training. \n
            See `RidgeRegression
            <https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.RidgeRegression.html>`_
            for more details.
        cv (Optional[sklearn.model_selection._split.BaseCrossValidator]):
            An Scikit-Learn cross-validator class.
            If not ``None``, then must have: \n
                * Call signature: ``idx_train, idx_test =
                  next(self.cv.split(self.X, self.y))`` \n
            If ``None``, then a ShuffleSplit cross-validator will be
            used.
        groups (Optional[np.ndarray]):
            Group labels for the samples used while splitting the dataset into
            train/test set. \n
        test_size (float):
            Test set ratio.
            Only used if ``cv`` is ``None``.
        loss_function (Optional[Callable]):
            A custom loss function. If ``None``, then the default loss function
            is used (LossFunction_MSE_CV).\n
        verbose (bool):
            Whether to print progress messages.
        optuna_storage_url (Optional[str]):
            The path or URL to a SQL database. If ``None``, then the study is
            stored in memory. See optuna documentation for details. \n
        optuna_storage_name (Optional[str]):
            The name of the study in the storage. Should only be ``None`` if
            ``optuna_storage_url`` is ``None``, else it should be a string to a
            valid study name. 
        optuna_engine_kwargs (Optional[Dict[str, Any]]):
            Additional keyword arguments to pass to the SQLAlchemy engine.
        wandb_project (Optional[str]):
            The name of the Weights and Biases project to log to. If ``None``,
            then no logging is done. Assumes that a wandb.init() has already
            been called. \n


    Demo:
        .. code-block:: python

            ## Initialize sklearn RidgeRegression class with NO TUNING. All
            parameters are fixed.
            autoregressor = Auto_Regression(
                X,
                y,
                Model=sklearn.linear_model.Ridge,
                params_dynamic={},
                params_static={
                    'alpha': 1e-14,
                    'solver': 'lbfgs',
                },
            )

            ## Initialize with TUNING 'alpha', 'solver', and 'positive'. 'tol' is fixed.
            autoregressor = Auto_Regression(
                X,
                y,
                Model=sklearn.linear_model.Ridge,
                params_dynamic={
                    'alpha': [1e-1, 1e3],
                    'solver': ['lbfgs', 'sag', 'saga'],
                    'positive': [True, False],
                },
                params_static={
                    'tol': 0.0001,
                },
            )
    """

    def __init__(
        self,
        X: np.ndarray,
        y: np.ndarray,
        Model: Callable,
        params_dynamic: Dict = {
            'alpha': [1e-1, 1e5],
            'fit_intercept': [True, False],
        },
        params_static: Dict = {
            'max_iter': 1000,
            'tol': 0.0001,
            'solver': 'auto',
        },
        params_log: Dict = {
            'alpha': True,
        },
        params_linked: Dict = {},
        n_startup: int = 15,
        kwargs_convergence: Dict = {
            'n_patience': 50,
            'tol_frac': 0.05,
            'max_trials': 150,
            'max_duration': 60*10,
        }, 
        n_repeats: int = 1,
        fn_reduce_repeats: Callable = np.median,
        n_jobs_optuna: int = 1,
        penalty_testTrainRatio: float = 1.0,
        sample_weight: Optional[List[float]] = None,
        cv: Optional[sklearn.model_selection._split.BaseCrossValidator] = None,
        groups: Optional[np.ndarray] = None,
        test_size: float = 0.3,
        loss_function: Optional[Callable] = None,
        verbose: bool = True,
        optuna_storage_url: Optional[str] = None,
        optuna_storage_name: Optional[str] = None,
        optuna_engine_kwargs: Optional[Dict[str, Any]] = None,
        wandb_project: Optional[str] = None,
    ) -> None:
        """
        Initializes Auto_Regression with the given parameters and data.
        """
        ## Prepare sample weights
        self.sample_weight = sample_weight

        ## Prepare the loss function
        if loss_function is None:
            self.fn_loss = LossFunction_MSE_CV(
                penalty_testTrainRatio=penalty_testTrainRatio,
            )
        else:
            self.fn_loss = loss_function

        ## Prepare the cross-validation
        self.cv = sklearn.model_selection.ShuffleSplit(
            n_splits=1,
            test_size=test_size,
        ) if cv is None else cv

        ## Prepare dynamic parameters for optuna
        params = _infer_params_types(params_dynamic, params_log)

        ## Prepare the model class
        self.model_class = functools.partial(
            Model,
            **params_static,
        )

        ## Initialize the Autotuner superclass
        super().__init__(
            model_class=self.model_class,
            X=X,
            y=y,
            params=params,
            params_linked=params_linked,
            n_startup=n_startup,
            kwargs_convergence=kwargs_convergence,
            n_repeats=n_repeats,
            fn_reduce_repeats=fn_reduce_repeats,
            n_jobs_optuna=n_jobs_optuna,
            cv=self.cv,
            groups=groups,
            fn_loss=self.fn_loss,
            catch_convergence_warnings=True,
            verbose=verbose,
            optuna_storage_url=optuna_storage_url,
            optuna_storage_name=optuna_storage_name,
            optuna_engine_kwargs=optuna_engine_kwargs,
            wandb_project=wandb_project,
        )


def _infer_params_types(
    params, 
    params_log={}, 
    log_default=True,
):
    """
    Infers the type of each parameter in params.
    RH 2023

    Args:
        params_dynamic (Dict):
            Dictionary of model class initialization parameters. 
            For each item in the dictionary if item is: \n
                * ``list``: The parameter is tuned. If the values are numbers,
                  then the list wil be the bounds [low, high] to search over. If
                  the values are strings, then the list will be the categorical
                  values to search over. Note that the type of the parameter
                  (int, float, str) will be used to determine the type of the
                  optuna suggest.
                * **not** a ``list``: The parameter is fixed to the given value. \n
        params_log (Dict):
            Dictionary of parameters that vary logarithmically: \n
                * If True: parameter is set to logaritmically varying. \n
                * If False: parameter is set to linearly varying. \n
                * If not specified: parameter is set to logaritmically varying.
                  \n
    
    Returns:
        (Dict)
            params_dynamic:
                Dictionary of parameters to be tuned. \n
                Formatted as: \n
                .. code-block:: python
                    {
                        'param1': {
                            'type': 'categorical' or 'real' or 'int',
                            'kwargs': {
                                'choices': [val1, val2, ...] if categorical,
                                'low': val1,
                                'high': val2,
                                'step': val3,
                                'log': True or False,
                            }
                        },
                        'param2': {
                            ...
                        },
                        ...
                    }                    
    """
    params_dynamic = {}
    for key, val in params.items():
        ## Dynamic parameters
        if isinstance(val, list):
            ## Categorical parameters
            if isinstance(val[0], str):
                params_dynamic[key] = {
                    'type': 'categorical', 
                    'kwargs': {}
                }
                params_dynamic[key]['kwargs']['choices'] = val
            ## Numerical parameters
            elif isinstance(val[0], (int, float)):
                assert len(val) >= 2, f'Parameter "{key}" must have at least 2 values.'
                assert all([isinstance(v, type(val[0])) for v in val]), f'Parameter "{key}" must have all values of the same type. Found types {[type(v) for v in val]}.'
                kwargs = ['low', 'high', 'step', 'log']
                params_dynamic[key] = {
                    'type': 'real' if isinstance(val[0], float) else 'int',
                    'kwargs': {kwargs[ii]: val for ii, val in enumerate(val)},
                }
                params_dynamic[key]['kwargs']['log'] = params_log.get(key, log_default)
            else:
                raise ValueError(f'Parameter "{key}" has type {type(val[0])}, which is not supported.')
        ## Static parameters
        else:
            raise ValueError(f'Parameter "{key}" must be a list. Found type {type(val)}.')
    return params_dynamic


from typing import Any, Union, Callable
import pickle
import json
import yaml
from pathlib import Path
import zipfile

from tqdm.auto import tqdm

from . import path_helpers


def prepare_path(path, mkdir=False, exist_ok=True):
    """
    Checks if a directory or filepath for validity for different
     purposes: saving, loading, etc.
    If exists:
        If exist_ok=True: all good
        If exist_ok=False: raises error
    If doesn't exist:
        If file:
            If parent directory exists:
                All good
            If parent directory doesn't exist:
                If mkdir=True: creates parent directory
                If mkdir=False: raises error
        If directory:
            If mkdir=True: creates directory
            If mkdir=False: raises error
   
    Returns a resolved path.
    RH 2023

    Args:
        path (str):
            Path to check.
        mkdir (bool):
            If True, creates parent directory if it does not exist.
        exist_ok (bool):
            If True, allows overwriting of existing file.

    Returns:
        path (str):
            Resolved path.
    """
    ## check if path is valid
    try:
        path_obj = Path(path).resolve()
    except FileNotFoundError as e:
        print(f'Invalid path: {path}')
        raise e
    
    ## check if path object exists
    flag_exists = path_obj.exists()

    ## determine if path is a directory or file
    if flag_exists:
        flag_dirFileNeither = 'dir' if path_obj.is_dir() else 'file' if path_obj.is_file() else 'neither'  ## 'neither' should never happen since path.is_file() or path.is_dir() should be True if path.exists()
        assert flag_dirFileNeither != 'neither', f'Path: {path} is neither a file nor a directory.'
        assert exist_ok, f'{path} already exists and exist_ok=False.'
    else:
        flag_dirFileNeither = 'dir' if path_obj.suffix == '' else 'file'  ## rely on suffix to determine if path is a file or directory

    ## if path exists and is a file or directory
    # all good. If exist_ok=False, then this should have already been caught above.
    
    ## if path doesn't exist and is a file
    ### if parent directory exists        
    # all good
    ### if parent directory doesn't exist
    #### mkdir if mkdir=True and raise error if mkdir=False
    if not flag_exists and flag_dirFileNeither == 'file':
        if Path(path).parent.exists():
            pass ## all good
        elif mkdir:
            Path(path).parent.mkdir(parents=True, exist_ok=True)
        else:
            assert False, f'File: {path} does not exist, Parent directory: {Path(path).parent} does not exist, and mkdir=False.'
        
    ## if path doesn't exist and is a directory
    ### mkdir if mkdir=True and raise error if mkdir=False
    if not flag_exists and flag_dirFileNeither == 'dir':
        if mkdir:
            Path(path).mkdir(parents=True, exist_ok=True)
        else:
            assert False, f'{path} does not exist and mkdir=False.'

    ## if path is neither a file nor a directory
    ### raise error
    if flag_dirFileNeither == 'neither':
        assert False, f'{path} is neither a file nor a directory. This should never happen. Check this function for bugs.'

    return str(path_obj)

### Custom functions for preparing paths for saving and loading files and directories
def prepare_filepath_for_saving(filepath, mkdir=False, allow_overwrite=False):
    return prepare_path(filepath, mkdir=mkdir, exist_ok=allow_overwrite)
def prepare_filepath_for_loading(filepath, must_exist=True):
    path = prepare_path(filepath, mkdir=False, exist_ok=must_exist)
    if must_exist:
        assert Path(path).is_file(), f'{path} is not a file.'
    return path
def prepare_directory_for_saving(directory, mkdir=False, exist_ok=True):
    """Rarely used."""
    return prepare_path(directory, mkdir=mkdir, exist_ok=exist_ok)
def prepare_directory_for_loading(directory, must_exist=True):
    """Rarely used."""
    path = prepare_path(directory, mkdir=False, exist_ok=must_exist)
    if must_exist:
        assert Path(path).is_dir(), f'{path} is not a directory.'
    return path


def pickle_save(
    obj, 
    filepath, 
    mode='wb', 
    zipCompress=False, 
    mkdir=False, 
    allow_overwrite=False,
    library='pickle',
    **kwargs_zipfile,
):
    """
    Saves an object to a pickle file.
    Allows for zipping of file.
    Uses pickle.dump.
    RH 2022

    Args:
        obj (object):
            Object to save.
        filepath (str):
            Path to save object to.
        mode (str):
            Mode to open file in.
            Can be:
                'wb' (write binary)
                'ab' (append binary)
                'xb' (exclusive write binary. Raises FileExistsError if file already exists.)
        zipCompress (bool):
            If True, compresses pickle file using zipfileCompressionMethod.
            This is similar to savez_compressed in numpy (with zipfile.ZIP_DEFLATED),
             and is useful for saving redundant and/or sparse arrays objects.
        mkdir (bool):
            If True, creates parent directory if it does not exist.
        allow_overwrite (bool):
            If True, allows overwriting of existing file.  
        library (str):
            Library to use for pickling. Can be:\n
                * 'pickle': Uses the built-in pickle library.\n
                * 'dill': Uses the dill library. Useful for pickling objects
                  that cannot be pickled with the built-in pickle library.\n
                * 'cloudpickle': Uses the cloudpickle library. Also very
                  flexible, but requires that the file is loaded on the same
                  version of Python that it was saved on.
        kwargs_zipfile (dict):
            Keyword arguments that will be passed into zipfile.ZipFile.
            compression=zipfile.ZIP_DEFLATED by default.
            See https://docs.python.org/3/library/zipfile.html#zipfile-objects.
            Other options for 'compression' are (input can be either int or object):
                0:  zipfile.ZIP_STORED (no compression)
                8:  zipfile.ZIP_DEFLATED (usual zip compression)
                12: zipfile.ZIP_BZIP2 (bzip2 compression) (usually not as good as ZIP_DEFLATED)
                14: zipfile.ZIP_LZMA (lzma compression) (usually better than ZIP_DEFLATED but slower)
    """
    pickle = __import__(library)
    path = prepare_filepath_for_saving(filepath, mkdir=mkdir, allow_overwrite=allow_overwrite)

    if len(kwargs_zipfile)==0:
        kwargs_zipfile = {
            'compression': zipfile.ZIP_DEFLATED,
        }

    if zipCompress:
        with zipfile.ZipFile(path, 'w', **kwargs_zipfile) as f:
            f.writestr('data', pickle.dumps(obj))
    else:
        with open(path, mode) as f:
            pickle.dump(obj, f)

def pickle_load(
    filepath, 
    zipCompressed=False,
    mode='rb',
    library='pickle',
):
    """
    Loads a pickle file.
    Allows for loading of zipped pickle files.
    RH 2022

    Args:
        filepath (str):
            Path to pickle file.
        zipCompressed (bool):
            If True, then file is assumed to be a .zip file.
            This function will first unzip the file, then
             load the object from the unzipped file.
        mode (str):
            Mode to open file in.
        library (str):
            Library to use for pickling. Can be:\n
                * 'pickle': Uses the built-in pickle library.\n
                * 'dill': Uses the dill library. Useful for pickling objects
                  that cannot be pickled with the built-in pickle library.\n
                * 'cloudpickle': Uses the cloudpickle library. Also very
                  flexible, but requires that the file is loaded on the same
                  version of Python that it was saved on.

    Returns:
        obj (object):
            Object loaded from pickle file.
    """
    pickle = __import__(library)
    path = prepare_filepath_for_loading(filepath, must_exist=True)
    if zipCompressed:
        with zipfile.ZipFile(path, 'r') as f:
            return pickle.loads(f.read('data'))
    else:
        with open(path, mode) as f:
            return pickle.load(f)


def json_save(obj, filepath, indent=4, mode='w', mkdir=False, allow_overwrite=False):
    """
    Saves an object to a json file.
    Uses json.dump.
    RH 2022

    Args:
        obj (object):
            Object to save.
        filepath (str):
            Path to save object to.
        mode (str):
            Mode to open file in.
            Can be:
                'wb' (write binary)
                'ab' (append binary)
                'xb' (exclusive write binary. Raises FileExistsError if file already exists.)
        mkdir (bool):
            If True, creates parent directory if it does not exist.
        allow_overwrite (bool):
            If True, allows overwriting of existing file.        
    """
    path = prepare_filepath_for_saving(filepath, mkdir=mkdir, allow_overwrite=allow_overwrite)
    with open(path, mode) as f:
        json.dump(obj, f, indent=indent)

def json_load(filepath, mode='r'):
    """
    Loads a json file.
    RH 2022

    Args:
        filepath (str):
            Path to json file.
        mode (str):
            Mode to open file in.

    Returns:
        obj (object):
            Object loaded from json file.
    """
    path = prepare_filepath_for_loading(filepath, must_exist=True)
    with open(path, mode) as f:
        return json.load(f)


def yaml_save(obj, filepath, indent=4, mode='w', mkdir=False, allow_overwrite=False):
    """
    Saves an object to a yaml file.
    Uses yaml.dump.
    RH 2022

    Args:
        obj (object):
            Object to save.
        filepath (str):
            Path to save object to.
        mode (str):
            Mode to open file in.
            Can be:
                'wb' (write binary)
                'ab' (append binary)
                'xb' (exclusive write binary. Raises FileExistsError if file already exists.)
        mkdir (bool):
            If True, creates parent directory if it does not exist.
        allow_overwrite (bool):
            If True, allows overwriting of existing file.        
    """
    path = prepare_filepath_for_saving(filepath, mkdir=mkdir, allow_overwrite=allow_overwrite)
    with open(path, mode) as f:
        yaml.dump(obj, f, indent=indent, sort_keys=False)

def yaml_load(filepath, mode='r', loader=yaml.FullLoader):
    """
    Loads a yaml file.
    RH 2022

    Args:
        filepath (str):
            Path to yaml file.
        mode (str):
            Mode to open file in.
        loader (yaml.Loader):
            Loader to use.
            Can be:
                yaml.FullLoader: Loads the full YAML language. Avoids arbitrary code execution. Default for PyYAML 5.1+.
                yaml.SafeLoader: Loads a subset of the YAML language, safely. This is recommended for loading untrusted input.
                yaml.UnsafeLoader: The original Loader code that could be easily exploitable by untrusted data input.
                yaml.BaseLoader: Only loads the most basic YAML. All scalars are loaded as strings.

    Returns:
        obj (object):
            Object loaded from yaml file.
    """
    path = prepare_filepath_for_loading(filepath, must_exist=True)
    with open(path, mode) as f:
        return yaml.load(f, Loader=loader)
    

def matlab_load(
    filepath, 
    simplify_cells=True, 
    kwargs_scipy={}, 
    kwargs_mat73={}, 
    verbose=False
):
    """
    Loads a matlab file.
    Uses scipy.io.loadmat if .mat file is not version 7.3.
    Uses mat73.loadmat if .mat file is version 7.3.
    RH 2023

    Args:
        filepath (str):
            Path to matlab file.
        simplify_cells (bool):
            If True and file is not version 7.3, then
             simplifies cells to numpy arrays.
        kwargs_scipy (dict):
            Keyword arguments to pass to scipy.io.loadmat.
        kwargs_mat73 (dict):
            Keyword arguments to pass to mat73.loadmat.
        verbose (bool):
            If True, prints information about file.
    """
    path = prepare_filepath_for_loading(filepath, must_exist=True)
    assert path.endswith('.mat'), 'File must be .mat file.'

    try:
        import scipy.io
        out = scipy.io.loadmat(path, simplify_cells=simplify_cells, **kwargs_scipy)
    except NotImplementedError as e:
        print(f'File {path} is version 7.3. Loading with mat73.') if verbose else None
        import mat73
        out = mat73.loadmat(path, **kwargs_mat73)
        print(f'Loaded {path} with mat73.') if verbose else None
    return out

def matlab_save(
    obj,
    filepath,
    mkdir=False, 
    allow_overwrite=False,
    clean_string=True,
    list_to_objArray=True,
    none_to_nan=True,
    kwargs_scipy_savemat={
        'appendmat': True,
        'format': '5',
        'long_field_names': False,
        'do_compression': False,
        'oned_as': 'row',
    }
):
    """
    Saves data to a matlab file.
    Uses scipy.io.savemat.
    Provides additional functionality by cleaning strings,
     converting lists to object arrays, and converting None to
     np.nan.
    RH 2023

    Args:
        obj (dict):
            Data to save.
        filepath (str):
            Path to save file to.
        clean_string (bool):
            If True, converts strings to bytes.
        list_to_objArray (bool):
            If True, converts lists to object arrays.
        none_to_nan (bool):
            If True, converts None to np.nan.
        kwargs_scipy_savemat (dict):
            Keyword arguments to pass to scipy.io.savemat.
    """
    import numpy as np

    prepare_filepath_for_saving(filepath, mkdir=mkdir, allow_overwrite=allow_overwrite)

    def walk(d, fn):
        return {key: fn(val) if isinstance(val, dict)==False else walk(val, fn) for key, val in d.items()}
    
    fn_clean_string = (lambda x: x.encode('utf-8') if isinstance(x, str) and clean_string else x) if clean_string else (lambda x: x)
    fn_list_to_objArray = (lambda x: np.array(x, dtype=object) if isinstance(x, list) and list_to_objArray else x) if list_to_objArray else (lambda x: x)
    fn_none_to_nan = (lambda x: np.nan if x is None and none_to_nan else x) if none_to_nan else (lambda x: x)

    data_cleaned = walk(walk(walk(obj, fn_clean_string), fn_list_to_objArray), fn_none_to_nan)

    import scipy.io
    scipy.io.savemat(filepath, data_cleaned, **kwargs_scipy_savemat)


# def zarr_save(
#     obj,
#     filepath,
#     mode='w',
#     mkdir=False,
#     allow_overwrite=False,
#     function_unsupported: Callable = lambda data: repr(data),
#     **kwargs_zarr,
# ):
#     """
#     Saves an object to a zarr file. Uses recursive approach to save
#     hierarchical objects.
#     RH 2024

#     Args:
#         obj (object):
#             Object to save. Can be any array or hierarchical object.
#         filepath (str):
#             Path to save object to.
#         mode (str):
#             Mode to open file in.
#             Can be:
#                 'wb' (write binary)
#                 'ab' (append binary)
#                 'xb' (exclusive write binary. Raises FileExistsError if file already exists.)
#         mkdir (bool):
#             If True, creates parent directory if it does not exist.
#         allow_overwrite (bool):
#             If True, allows overwriting of existing file.        
#         kwargs_zarr (dict):
#             Keyword arguments to pass to zarr.save.
#     """
#     import zarr

#     path = prepare_filepath_for_saving(filepath, mkdir=mkdir, allow_overwrite=allow_overwrite)
    
#     import numpy as np
#     import scipy.sparse

#     def save_data_to_zarr(data, group, name=None):
#         """
#         Recursively saves complex nested data structures into a Zarr group.

#         Parameters:
#         - data: The data to save (dict, list, tuple, np.ndarray, int, float, str, bool, or None).
#         - group: The Zarr group to save data into.
#         - name: The name of the dataset or subgroup (used in recursive calls).
#         """
#         if isinstance(data, dict):
#             # Use the given name or the current group
#             sub_group = group.require_group(name) if name else group
#             for key, value in data.items():
#                 # Ensure keys are strings
#                 key_str = str(key) if not isinstance(key, str) else key
#                 # Recursively save data
#                 save_data_to_zarr(value, sub_group, name=key_str)
#         elif isinstance(data, (list, tuple)):
#             # Create a subgroup for lists and tuples
#             sub_group = group.require_group(name) if name else group
#             for idx, item in enumerate(data):
#                 key_str = str(idx)
#                 save_data_to_zarr(item, sub_group, name=key_str)
#         elif isinstance(data, np.ndarray):
#             if name is None:
#                 raise ValueError("Name must be provided for dataset")
#             group.create_dataset(name, data=data)
#         elif isinstance(data, scipy.sparse.spmatrix):
#             if name is None:
#                 raise ValueError("Name must be provided for dataset")
#             group.create_dataset(name, data=data)
#         elif isinstance(data, (int, float, str, bool)):
#             if name is None:
#                 raise ValueError("Name must be provided for dataset")
#             group.create_dataset(name, data=data)
#         elif data is None:
#             if name is None:
#                 raise ValueError("Name must be provided for dataset")
#             # Store None as a special attribute
#             group.attrs[name] = 'None'
#         else:
#             # For unsupported types, store the string representation
#             if name is None:
#                 raise ValueError("Name must be provided for attribute")
#             group.attrs[name] = repr(data)

#     zarr_group = zarr.open(path, mode=mode, **kwargs_zarr)
#     save_data_to_zarr(obj, zarr_group, name=None)


# def zarr_load(
#     filepath,
#     mode='r',
#     **kwargs_zarr,
# ):
#     """
#     Loads a zarr file. Uses recursive approach to load hierarchical
#     objects.
#     RH 2024

#     Args:
#         filepath (str):
#             Path to zarr file.
#         mode (str):
#             Mode to open file in.
#         kwargs_zarr (dict):
#             Keyword arguments to pass to zarr.load.
#     """
#     import zarr

#     # path = prepare_filepath_for_loading(filepath, must_exist=True)
#     path = filepath
    
#     def load_data_from_zarr(group):
#         """
#         Recursively loads complex nested data structures from a Zarr group.

#         Parameters:
#         - group: The Zarr group to load data from.

#         Returns:
#         - The loaded data (dict, list, tuple, np.ndarray, int, float, str, bool, or None).
#         """
#         data = {}
#         for key in group.array_keys():
#             data[key] = group[key][...]
#         for key in group.group_keys():
#             data[key] = load_data_from_zarr(group[key])
#         for key, value in group.attrs.items():
#             if value == 'None':
#                 data[key] = None
#             else:
#                 try:
#                     data[key] = eval(value)
#                 except Exception:
#                     data[key] = value
#         return data

#     zarr_group = zarr.open(path, mode=mode, **kwargs_zarr)
#     return load_data_from_zarr(zarr_group)
        

def hash_file(path, type_hash='MD5', buffer_size=65536):
    """
    Gets hash of a file.
    Based on: https://stackoverflow.com/questions/22058048/hashing-a-file-in-python
    RH 2022

    Args:
        path (str):
            Path to file to be hashed.
        type_hash (str):
            Type of hash to use. Can be:
                'MD5'
                'SHA1'
                'SHA256'
                'SHA512'
        buffer_size (int):
            Buffer size for reading file.
            65536 corresponds to 64KB.

    Returns:
        hash_val (str):
            Hash of file.
    """
    import hashlib

    if type_hash == 'MD5':
        hasher = hashlib.md5()
    elif type_hash == 'SHA1':
        hasher = hashlib.sha1()
    elif type_hash == 'SHA256':
        hasher = hashlib.sha256()
    elif type_hash == 'SHA512':
        hasher = hashlib.sha512()
    else:
        raise ValueError(f'{type_hash} is not a valid hash type.')

    with open(path, 'rb') as f:
        while True:
            data = f.read(buffer_size)
            if not data:
                break
            hasher.update(data)

    hash_val = hasher.hexdigest()
        
    return hash_val


def download_file(
    url, 
    path_save, 
    check_local_first=True, 
    check_hash=False, 
    hash_type='MD5', 
    hash_hex=None,
    mkdir=False,
    allow_overwrite=False,
    write_mode='wb',
    verbose=True,
    chunk_size=1024,
):
    """
    Download a file from a URL to a local path using requests.
    Allows for checking if file already exists locally and
    checking the hash of the downloaded file against a provided hash.
    RH 2022

    Args:
        url (str):
            URL of file to download.
            If url is None, then no download is attempted.
        path_save (str):
            Path to save file to.
        check_local_first (bool):
            If True, checks if file already exists locally.
            If True and file exists locally, plans to skip download.
            If True and check_hash is True, checks hash of local file.
             If hash matches, skips download. If hash does not match, 
             downloads file.
        check_hash (bool):
            If True, checks hash of local or downloaded file against
             hash_hex.
        hash_type (str):
            Type of hash to use. Can be:
                'MD5', 'SHA1', 'SHA256', 'SHA512'
        hash_hex (str):
            Hash to compare to. In hex format (e.g. 'a1b2c3d4e5f6...').
            Can be generated using hash_file() or hashlib and .hexdigest().
            If check_hash is True, hash_hex must be provided.
        mkdir (bool):
            If True, creates parent directory of path_save if it does not exist.
        write_mode (str):
            Write mode for saving file. Should be one of:
                'wb' (write binary)
                'ab' (append binary)
                'xb' (write binary, fail if file exists)
        verbose (bool):
            If True, prints status messages.
        chunk_size (int):
            Size of chunks to download file in.
    """
    import os
    import requests

    path_save = prepare_filepath_for_saving(path_save, mkdir=mkdir, allow_overwrite=allow_overwrite)

    # Check if file already exists locally
    if check_local_first:
        if os.path.isfile(path_save):
            print(f'File already exists locally: {path_save}') if verbose else None
            # Check hash of local file
            if check_hash:
                hash_local = hash_file(path_save, type_hash=hash_type)
                if hash_local == hash_hex:
                    print('Hash of local file matches provided hash_hex.') if verbose else None
                    return True
                else:
                    print('Hash of local file does not match provided hash_hex.') if verbose else None
                    print(f'Hash of local file: {hash_local}') if verbose else None
                    print(f'Hash provided in hash_hex: {hash_hex}') if verbose else None
                    print('Downloading file...') if verbose else None
            else:
                return True
        else:
            print(f'File does not exist locally: {path_save}. Will attempt download from {url}') if verbose else None

    # Download file
    if url is None:
        print('No URL provided. No download attempted.') if verbose else None
        return None
    try:
        response = requests.get(url, stream=True)
    except requests.exceptions.RequestException as e:
        print(f'Error downloading file: {e}') if verbose else None
        return False
    # Check response
    if response.status_code != 200:
        print(f'Error downloading file. Response status code: {response.status_code}') if verbose else None
        return False
    # Create parent directory if it does not exist
    prepare_filepath_for_saving(path_save, mkdir=mkdir, allow_overwrite=allow_overwrite)
    # Download file with progress bar
    total_size = int(response.headers.get('content-length', 0))
    wrote = 0
    with open(path_save, write_mode) as f:
        with tqdm(total=total_size, disable=(verbose==False), unit='B', unit_scale=True, unit_divisor=1024) as pbar:
            for data in response.iter_content(chunk_size):
                wrote = wrote + len(data)
                f.write(data)
                pbar.update(len(data))
    if total_size != 0 and wrote != total_size:
        print("Error downloading file. Downloaded file is not the same size as the file on the server.")
        return False
    # Check hash
    if check_hash:
        hash_local = hash_file(path_save, type_hash=hash_type)
        if hash_local == hash_hex:
            print('Hash of downloaded file matches hash_hex.') if verbose else None
            return True
        else:
            print('Hash of downloaded file does not match hash_hex.') if verbose else None
            print(f'Hash of downloaded file: {hash_local}') if verbose else None
            print(f'Hash provided in hash_hex: {hash_hex}') if verbose else None
            return False
    else:
        return True

def is_valid_hash(hash_hex, hash_type='MD5'):
    """
    Checks if a hash is valid.
    RH 2022

    Args:
        hash_hex (str):
            Hash to check. In hex format (e.g. 'a1b2c3d4e5f6...').
        hash_type (str):
            Type of hash to use. Can be:
                'MD5', 'SHA1', 'SHA256', 'SHA512'

    Returns:
        bool:
            True if hash is valid, False if not.
    """
    import hashlib

    if hash_type == 'MD5':
        hasher = hashlib.md5()
    elif hash_type == 'SHA1':
        hasher = hashlib.sha1()
    elif hash_type == 'SHA256':
        hasher = hashlib.sha256()
    elif hash_type == 'SHA512':
        hasher = hashlib.sha512()
    else:
        raise ValueError(f'{hash_type} is not a valid hash type.')

    try:
        hasher.update(bytes.fromhex(hash_hex))
    except ValueError:
        return False
    return True


def compare_file_hashes(
    hash_dict_true,
    dir_files_test=None,
    paths_files_test=None,
    verbose=True,
):
    """
    Compares hashes of files in a directory or list of paths
     to user provided hashes.
    RH 2022

    Args:
        hash_dict_true (dict):
            Dictionary of hashes to compare to.
            Each entry should be:
                {'key': ('filename', 'hash')}
        dir_files_test (str):
            Path to directory to compare hashes of files in.
            Unused if paths_files_test is not None.
        paths_files_test (list of str):
            List of paths to files to compare hashes of.
            Optional. dir_files_test is used if None.
        verbose (bool):
            Whether or not to print out failed comparisons.

    Returns:
        total_result (bool):
            Whether or not all hashes were matched.
        individual_results (list of bool):
            Whether or not each hash was matched.
        paths_matching (dict):
            Dictionary of paths that matched.
            Each entry is:
                {'key': 'path'}
    """
    if paths_files_test is None:
        if dir_files_test is None:
            raise ValueError('Must provide either dir_files_test or path_files_test.')
        
        ## make a dict of {filename: path} for each file in dir_files_test
        files_test = {filename: (Path(dir_files_test).resolve() / filename).as_posix() for filename in path_helpers.get_dir_contents(dir_files_test)[1]} 
    
    paths_matching = {}
    results_matching = {}
    for key, (filename, hash) in hash_dict_true.items():
        match = True
        if filename not in files_test:
            print(f'{filename} not found in test directory: {dir_files_test}.') if verbose else None
            match = False
        elif hash != hash_file(files_test[filename]):
            print(f'{filename} hash mismatch with {key, filename}.') if verbose else None
            match = False
        if match:
            paths_matching[key] = files_test[filename]
        results_matching[key] = match

    return all(results_matching.values()), results_matching, paths_matching


def extract_zip(
    path_zipFile,
    path_extract=None,
    mkdir=True,
    verbose=True,
):
    """
    Extracts a zip file.
    RH 2022

    Args:
        path_zipFile (str):
            Path to zip file.
        path_extract (str):
            Path to extract zip file to.
            If None, extracts to the same directory as the zip file.
        verbose (int):
            Whether to print progress.
    """
    path_zipFile = prepare_filepath_for_loading(path_zipFile, must_exist=True)
    import zipfile
    if path_extract is None:
        path_extract = str(Path(path_zipFile).parent)
    path_extract = prepare_directory_for_saving(path_extract, mkdir=mkdir, exist_ok=True)

    print(f'Extracting {path_zipFile} to {path_extract}.') if verbose else None

    with zipfile.ZipFile(path_zipFile, 'r') as zip_ref:
        zip_ref.extractall(path_extract)

    print('Completed zip extraction.') if verbose else None
